/*
 * SynchroList.java - Synchronized List
 *
 * Copyright (C) 1998 Charles McManis, All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 * CHUCK MCMANIS MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
 * SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. CHUCK MCMANIS
 * SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT
 * OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 */

import java.util.Enumeration;

/**
 * The Synchronized List
 *
 * A synchronized list is one in which there can be multiple threads
 * of execution putting things on the list, and enumerating the list.
 *
 */
public class SynchroList {

    /**
     * The inner Link class is used to link the objects together into
     * a doubly linked list. This "helper" class wraps both next and
     * previous pointers with a state vector to identify when a particular
     * object is currently participating in a list.
     */
    class Link {
        private Object data;
        private Link nxt, prv;

        /**
         * Create a new Link and insert it between nodes p and n
         */
        Link(Object o, Link p, Link n) {
            nxt = n; prv = p; data = o;
            if (n != null)
                n.prv = this;
            if (p != null)
                p.nxt = this;
        }

        /** return the held object reference */
        Object getData() { return data; }

        /** return the link after this one. */
        Link next() { return nxt; }
        Link next(Link newNext) { Link r = nxt; nxt = newNext; return r;}

        /** return the link before this one. */
        Link prev() { return prv; }
        Link prev(Link newPrev) { Link r = prv; prv = newPrev; return r;}

        public String toString() { return "Link("+data+")"; }

    }

    /**
     * The inner Link Enumerator class provides a selected series
     * of objects from the pile. These can either be all of the objects
     * or a subset based on bounding from/to object pairs.
     */
    class LinkEnumerator implements Enumeration {
        private Link current;

        /**
         * Create an enumerator for all objects.
         */
        LinkEnumerator( ) {
            current = head;
        }

        /**
         * Implementation of the Enumeration Interface's hasMoreElements.
         * In this case if current in non-null then it will be returned on
         * the next call so we return true.
         */
        public boolean hasMoreElements() { return (current != null); }

        /**
         * Implementation of the Enumeration Interface's nextElement method.
         * We return the object from the current link, and note it for later.
         * when we return the next link we release our hold on the previous
         * one.
         */
        public Object nextElement() {
            Object result = null;

            if (current != null) {
                result = current.getData();
                current = current.next();
            }
            return result;
        }

    }


    Comparator cmp;
    Link head, tail;

    /**
     * Create an unsorted list.
     */
    public SynchroList() {  }

    /**
     * Create a prioritized list.
     */
    public SynchroList(Comparator c) { cmp = c; }

    /**
     * Insert the object before the pointed to Link.
     */
    private void before(Object o, Link p) {
        new Link(o, p.prev(), p);
    }

    /**
     * Insert the object after the pointed to link.
     */
    private void after(Object o, Link p) {
        new Link(o, p, p.next());
    }

    /**
     * Remove the Link indicated.
     */
    private void remove(Link p) {
        if (p.prev() == null) {
            head = p.next();
            (p.next()).prev(null);
        } else if (p.next() == null) {
            tail = p.prev();
            (p.prev()).next(null);
        } else {
            p.prev().next(p.next());
            p.next().prev(p.prev());
        }
    }

    /**
     * Interface for adding objects to the collection. In the example you
     * cannot delete objects, although adding deletion is pretty straight
     * forward.
     */
    public void add(Object o) {
        // if cmp is null, always add to the tail of the list.
        if (cmp == null) {
            if (head == null) {
                head = new Link(o, null, null);
                tail = head;
            } else {
                tail = new Link(o, tail, null);
            }
            return;
        }

        cmp.typeCheck(o); // Validate its type

        if (head == null) {
            head = new Link(o, null, null);
            tail = head;
        } else if (cmp.lessThan(o, head.getData())) {
           head = new Link(o, null, head);
        } else {
            Link l;
            for (l = head; l.next() != null; l = l.next()) {
                if (cmp.lessThan(o, l.getData())) {
                    before(o, l);
                    return;
                }
            }
            tail = new Link(o, tail, null);
        }
        return;
    }

    public boolean delete(Object o) {
        if (cmp == null)
            return false;
        for (Link l = head; l != null; l = l.next()) {
            if (cmp.equalTo(o, l.getData())) {
               remove(l);
               return true;
            }
            if (cmp.lessThan(o, l.getData()))
               break;
        }
        return false;
    }


    /**
     * The enumerator generators are synchronized because the list must
     * remain stable during their generation, however once generated
     * the object is no longer synchronized. The ListEnumerator object will
     * prevent corruption to the list.
     */
    public synchronized Enumeration elements() {
        return new LinkEnumerator();
    }

    public int size() {
        int result = 0;
        for (Link l = head; l != null; l = l.next()) result++;
        return result;
    }
}
